/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.osgi.launcher.core;

import java.io.File;
import java.io.FileInputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.repository.DefaultRepositoryRequest;
import org.apache.maven.artifact.repository.RepositoryRequest;
import org.apache.maven.artifact.resolver.ArtifactResolutionRequest;
import org.apache.maven.artifact.resolver.ArtifactResolutionResult;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.model.Dependency;
import org.apache.maven.model.DependencyManagement;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugins.annotations.Component;
import org.apache.maven.plugins.annotations.Parameter;
import org.apache.maven.project.MavenProject;
import org.apache.maven.repository.RepositorySystem;
import org.ow2.util.maven.osgi.launcher.api.ILauncher;
import org.ow2.util.url.URLUtils;


/**
 * Common class for the Mojo (start or stop) and defines the common OSGi stuff.
 * @author Florent BENOIT
 */
public abstract class AbsMojo extends AbstractMojo {

    /**
     * Key for keeping the launcher (local or remote).
     */
    private static final String LAUNCHER_KEY = "EasyBeans-OSGi-Launcher";

    /**
     * Bundle file extension array.
     */
    private static final String[] BUNDLE_FILE_EXTENSIONS = new String[] {".jar"};

    /**
     * Path of the build directory.
     * @since 2.0.0
     */
    @Parameter(defaultValue = "${project.build.directory}")
    private File buildDirectory;

    @Component
    protected MavenSession mavenSession;

    @Component
    protected RepositorySystem repositorySystem;

    /**
     * The maven project.
     */
    @Component
    private MavenProject project;

    /**
     * List of directories that contains bundles.
     * @since 2.0.0
     */
    @Parameter
    private File[] bundleDirectories = null;

    /**
     * Fork the JVM for the start of the framework ? (default is false)
     * @since 2.0.0
     */
    @Parameter(defaultValue = "false")
    protected boolean fork = false;

    /**
     * Debug the forked JVM for the start of the framework ? (default is false)
     * @since 2.0.0
     */
    @Parameter
    protected Debug debug = new Debug();

    /**
     * List of Modules that needs to be installed/started.
     * @since 2.0.0
     */
    @Parameter
    private Module[] modules;

    /**
     * List of artifacts used as dependencies for the launcher.
     * @since 2.0.0
     */
    @Parameter
    private org.ow2.util.maven.osgi.launcher.core.Dependency[] frameworkDependencies;

    /**
     * Configuration file.
     * @since 2.0.0
     */
    @Parameter
    private File configurationFile = null;

    /**
     * Bundle configuration file (define bootstrap bundles).
     * @since 2.0.0
     */
    @Parameter
    private File bundleConfigurationFile = null;

    /**
     * An ordered list of Properties files containing elements of the configuration.
     * They will be merged and resolved together to obtain the final configuration.
     * @since 2.0.0
     */
    @Parameter
    private File[] configurationsFiles;

    /**
     * JVM Properties (System properties to set).
     * @since 2.0.0
     */
    @Parameter
    private Properties jvmProperties = new Properties();

    /**
     * List of bundles to stop in the given order (by their symbolic name).
     * @since 2.0.0
     */
    @Parameter
    private String[] stopOrderBundles = null;

    /**
     * List of bundles to ignore (by their symbolic name).
     * @since 2.0.0
     */
    @Parameter
    private String[] excludedBundles = null;

    /**
     * List of path for the user Classpath.
     * @since 2.0.0
     */
    @Parameter
    private String[] userClasspath = null;


    /**
     * List of bundles to start.
     */
    private List<URL> bundles = null;

    /**
     * List of Artifacts to Start.
     */
    private List<Artifact> artifacts = null;

    /**
     * List of Artifacts to add as dependencies.
     */
    private List<Artifact> frameWorkDependenciesArtifacts = null;

    /**
     * Waiting value after the stop.
     * @since 2.0.0
     */
    @Parameter(defaultValue = "0")
    private long waitAfterStop = 0;


    /**
     * Configuration.
     */
    private Configuration configuration = null;

    /**
     * Default constructor.
     */
    public AbsMojo() {
        this.bundles = new LinkedList<URL>();
        this.artifacts = new LinkedList<Artifact>();
        this.frameWorkDependenciesArtifacts = new LinkedList<Artifact>();
        this.configuration = new Configuration();
    }

    /**
     * Execute the Maven plugin.
     * @throws MojoExecutionException if the file is not generated.
     */
    public void execute() throws MojoExecutionException {

        // Add bootstrap artifacts
        resolveBootstrapArtifacts();

        // Add artifacts given by the user
        resolveArtifacts(modules, this.artifacts);
        resolveArtifacts(frameworkDependencies, this.frameWorkDependenciesArtifacts);

        extractBundlesFromArtifacts();

        // Extract bundles from the given directories
        extractBundlesFromDirectories();

        doExecute();
    }

    /**
     * Subclass execution (code for start, stop, etc).
     * @throws MojoExecutionException if the execution fails
     */
    protected abstract void doExecute() throws MojoExecutionException;

    /**
     * Extract the bundles from the given directories.
     */
    protected void extractBundlesFromDirectories() {
        // No directories, do nothing
        if (bundleDirectories == null) {
            return;
        }

        // Take bundles of the directories
        for (File directory : bundleDirectories) {
            addBundleFromDirectory(directory);
        }

    }

    /**
     * Extract the bundles from the given artifacts.
     */
    protected void extractBundlesFromArtifacts() {
        // Use URL of artifacts
        for (Artifact artifact : artifacts) {
            bundles.add(URLUtils.fileToURL(artifact.getFile()));
        }
    }

    /**
     * Add the bundles that can be found in the given directory.
     * @param directory the directory to scan
     */
    protected void addBundleFromDirectory(final File directory) {
        if (directory.isDirectory()) {
            File[] files = directory.listFiles();
            for (File f : files) {
                if (f.isDirectory()) {
                    addBundleFromDirectory(f);
                } else {
                    String fileName = f.getName().toLowerCase();
                    boolean validExtension = false;
                    int i = 0;
                    while (!validExtension && i < BUNDLE_FILE_EXTENSIONS.length) {
                        if (fileName.endsWith(BUNDLE_FILE_EXTENSIONS[i])) {
                            validExtension = true;
                        }
                        i++;
                    }
                    if (validExtension) {
                        bundles.add(URLUtils.fileToURL(f));
                    }
                }
            }
        }
    }

    /**
     * Gets a launcher if a previous one is existing or creates a new one if it is permitted.
     * @return a launcher (existing or newly created)
     * @throws MojoExecutionException if no launcher can't be found/created
     */
    @SuppressWarnings("unchecked")
    protected ILauncher getLauncher() throws MojoExecutionException {
        Map pluginContext = getPluginContext();

        ILauncher launcher = null;

        // Check context ?
        if (pluginContext != null) {
            // Existing launcher ?
            launcher = (ILauncher) pluginContext.get(getLauncherKey());
            if (launcher != null) {
                getLog().debug("Found a previous launcher in the plugin context");
            }
        }

        if (launcher == null) {
            launcher = createLauncher();
            pluginContext.put(getLauncherKey(), launcher);
        }

        return launcher;
    }


    /**
     * Build a new launcher (local or remote) depending of the fork mode.
     * @return a launcher.
     * @throws MojoExecutionException if no launcher is created
     */
    protected ILauncher createLauncher() throws MojoExecutionException {
        ILauncher launcher = null;

        // ForK ? create a remote launcher
        if (fork) {
            // Extract URLs of the current classloader

            // Get Method
            ClassLoader cl = Thread.currentThread().getContextClassLoader();
            Method getURLsMethod = null;
            try {
                getURLsMethod = URLClassLoader.class.getDeclaredMethod("getURLs");
            } catch (SecurityException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } catch (NoSuchMethodException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            }

            // Invoke method
            URL[] urls = null;
            boolean previousAccessible = getURLsMethod.isAccessible();
            getURLsMethod.setAccessible(true);
            try {
                urls = (URL[]) getURLsMethod.invoke(cl);
            } catch (IllegalArgumentException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } catch (IllegalAccessException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } catch (InvocationTargetException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } finally {
                getURLsMethod.setAccessible(previousAccessible);
            }

            List<URL> urlList = new LinkedList<URL>(Arrays.asList(urls));
            // add framework dependencies (if any)
            for (Artifact dependencyArtifact : frameWorkDependenciesArtifacts) {
                urlList.add(URLUtils.fileToURL(dependencyArtifact.getFile()));
            }

            getLog().debug("URLs for remote launcher are : " + urlList);
            launcher = new RemoteLauncher(urlList.toArray(new URL[urlList.size()]), getConfiguration(), debug);
        } else {
            launcher = new LocalLauncher(getConfiguration());
        }

        return launcher;
    }


    /**
     * Gets the bundles.
     * @return list of bundles
     */
    protected List<URL> getBundles() {
        return bundles;
    }

    /**
     * Found artifacts for bootstrap bundles.
     * @throws MojoExecutionException if artifact resolution fails
     */
    protected void resolveBootstrapArtifacts() throws MojoExecutionException {
        // No configuration file, break here
        if (bundleConfigurationFile == null) {
            return;
        }

        // Load properties of the 'jonas-autodeploy-bundles.properties' file
        Properties properties = new Properties();
        try {
            properties.load(new FileInputStream(bundleConfigurationFile));
        } catch (Exception e) {
            throw new MojoExecutionException("Fail to load bundle configuration file", e);
        }

        // Parse bootstrap bundles
        String[] bundleArtifacts = properties.getProperty("start.level.1").split(",");
        for (String bundleArtifact : bundleArtifacts) {
            String[] artifactValues = bundleArtifact.split(":");

            String groupId = artifactValues[0].trim();
            String artifactId = artifactValues[1];

            // Exclude gui bundles
            if (artifactId.startsWith("org.apache.felix.shell.gui")) {
                continue;
            }

            String type = "jar";
            String version = artifactValues[2];
            String classifier = null;
            if (artifactValues.length == 4) {
                classifier = artifactValues[3].replaceAll("\\{|\\}", "");
            }

            addArtifact(groupId, artifactId, version, type, classifier, artifacts);
        }
    }



    /**
     * Found artifacts for the list of the modules specified by the user.
     * @param modules the module to add to the given artifactList
     * @param artifactList the list to complete
     * @throws MojoExecutionException if artifact resolution fails
     */
    @SuppressWarnings("unchecked")
    protected void resolveArtifacts(final Module[] modules, final List<Artifact> artifactList) throws MojoExecutionException {
        // No module, break here
        if (modules == null) {
            return;
        }

        // for each module, build artifact.
        for (Module module : modules) {
            String groupId = module.getGroupId();
            String artifactId = module.getArtifactId();
            String type = module.getType();
            String version = module.getVersion();
            String classifier = module.getClassifier();

            if (groupId == null) {
                throw new MojoExecutionException("No groupdId set for module '" + module + "'");
            }
            if (artifactId == null) {
                throw new MojoExecutionException("No artifactId set for module '" + module + "'");
            }

            if (type == null) {
                getLog().debug("No type set for module '" + module + "', assume it is jar type");
                type = "jar";
                module.setType(type);
            }

            // No version
            if (version == null) {
                // try to find version with dependency management
                DependencyManagement dependencyManagement = project.getDependencyManagement();

                if (dependencyManagement != null) {
                    // try to find a version for this component
                    List<Dependency> dependencies = dependencyManagement.getDependencies();
                    if (dependencies != null) {
                        for (Dependency dependency : dependencies) {
                            getLog().debug("Dependency = " + dependency);
                            if (groupId.equals(dependency.getGroupId())
                                    && artifactId.equals(dependency.getArtifactId())
                                    && type.equals(dependency.getType())) {
                                getLog().debug("Using version of dependency management" + dependency.getVersion());
                                version = dependency.getVersion();
                                break;
                            }
                        }
                    }
                }
                if (version == null) {
                    throw new MojoExecutionException("No version set for module '" + module + "'");
                }
            }

            addArtifact(groupId, artifactId, version, type, classifier, artifactList);
        }
    }


    /**
     * Add artifact.
     * @param groupId artifact's groupId
     * @param artifactId artifact's artifactId
     * @param version artifact's version
     * @param type artifact's type
     * @param classifier artifact's classifier
     * @param artifacts the list on which to add the given artifact
     * @return the given artifact
     * @throws MojoExecutionException if artifact resolution fails
     */
    protected Artifact addArtifact(final String groupId, final String artifactId, final String version, final String type,
            final String classifier, final List<Artifact> artifacts) throws MojoExecutionException {
        Artifact artifact = repositorySystem.createArtifactWithClassifier(groupId, artifactId, version, type, classifier);
        getLog().debug("Built Artifact = " + artifact);

        // now resolve this artifact
        Artifact resolved = resolveArtifact(artifact);

        // Add artifact
        if (artifacts != null) {
            artifacts.add(resolved);
        }
        return artifact;
    }

    private Artifact resolveArtifact(Artifact artifact) throws MojoExecutionException {

        RepositoryRequest repositoryRequest = DefaultRepositoryRequest.getRepositoryRequest(mavenSession, project);
        ArtifactResolutionRequest request = new ArtifactResolutionRequest(repositoryRequest);
        request.setArtifact(artifact);

        // Resolve
        ArtifactResolutionResult result = repositorySystem.resolve(request);
        if (!result.isSuccess()) {
            for (Exception e : result.getExceptions()) {
                getLog().error(e);
            }

            throw new MojoExecutionException("Couldn't resolve artifact " + artifact);
        }
        return result.getArtifacts().iterator().next();
    }

    /**
     * @return the buildDirectory
     */
    public File getBuildDirectory() {
        return buildDirectory;
    }

    /**
     * @param buildDirectory the buildDirectory to set
     */
    public void setBuildDirectory(final File buildDirectory) {
        this.buildDirectory = buildDirectory;
    }

    /**
     * @return the modules
     */
    public Module[] getModules() {
        if (modules != null) {
            return modules.clone();
        }
        return null;
    }

    /**
     * @param modules the modules to set
     */
    public void setModules(final Module[] modules) {
        if (modules != null) {
            this.modules = modules.clone();
        }
    }

    /**
     * @return the jvmProperties
     */
    public Properties getJvmProperties() {
        return jvmProperties;
    }

    /**
     * @param jvmProperties the jvmProperties to set
     */
    public void setJvmProperties(final Properties jvmProperties) {
        this.jvmProperties = jvmProperties;
    }

    /**
     * Gets the configuration used by the launcher/manager.
     * @return a configuration object
     * @throws MojoExecutionException if the configuration can't be built.
     */
    protected Configuration getConfiguration() throws MojoExecutionException {

        // Time to wait after the stop
        configuration.setWaitAfterStop(waitAfterStop);

        // Stop bundles in a given order
        configuration.setStopOrderBundles(stopOrderBundles);

        // Ignore bundles
        configuration.setExcludedBundles(excludedBundles);

        // User classpath
        configuration.setUserClasspath(userClasspath);

        // JVM properties
        configuration.setSystemProperties(jvmProperties);

        // Bundles to deploy
        configuration.setBundles(bundles);

        // Set working directory
        configuration.setWorkingDirectory(new File(buildDirectory, "osgilauncher"));

        // Set Test directory
        configuration.setTestDirectory(new File(buildDirectory, "tests-report"));


        // Config file URL
        if (configurationFile != null) {
            configuration.setURLConfigFile(URLUtils.fileToURL(configurationFile));
        }

        // Configuration elements
        if (configurationsFiles != null && !(configurationsFiles.length == 0)) {
            configuration.setConfigurationElements(configurationsFiles);
        }
        return configuration;
    }


    /**
     * @return Launcher Key
     */
    protected String getLauncherKey() {
        return LAUNCHER_KEY;
    }



}
