/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.osgi.launcher.core;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.Enumeration;
import java.util.Properties;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * @author Florent BENOIT
 */
public final class Bootstrap {

    /**
     * Seconds is 1000 millis.
     */
    private static final long SECONDS = 1000L;

    /**
     * Two minutes are 120 seconds.
     */
    private static final long TWO_MINUTES = 120 * SECONDS;

    /**
     * Pause while analyzing the monitoring file.
     */
    private static final long WAIT_MONITORING = 500L;

    /**
     * Name of the configuration argument.
     */
    public static final String CONFIGURATION_PATH = "bootstrap.configuration.path";

    /**
     * Name of the configuration argument.
     */
    public static final String MONITORED_FILE_KEY = "bootstrap.monitored.file";

    /**
     * Name of the configuration argument.
     */
    public static final String KILL_JVM_TIMEOUT = "bootstrap.kill.timeout";

    /**
     * File that is used to receive calls for the shutdown.
     */
    public static final String MONITORED_FILE_PREFIX = System.getProperty("java.io.tmpdir") + File.separator
    + System.getProperty("user.name") + File.separator;

    /**
     * File that is used to receive calls for the shutdown.
     */
    public static final String MONITORED_FILE = MONITORED_FILE_PREFIX + "Launcher.lock";

    /**
     * Started instruction.
     */
    public static final String STARTED_INSTRUCTION = "STARTED";

    /**
     * Stopped instruction.
     */
    public static final String STOPPED_INSTRUCTION = "STOPPED";

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(Bootstrap.class);

    /**
     * Configuration.
     */
    private Configuration configuration = null;

    /**
     * Framework manager that will be used to start/stop framework.
     */
    private FrameworkManager frameworkManager = null;

    /**
     * Already called shutdown method ? if yes, avoid to do it again.
     */
    private boolean shutdownDone = false;

    /**
     * Monitored file set by user.
     */
    private static String  monitoredFile = MONITORED_FILE;

    /**
     * Time to wait after what, the jvm is killed whatever happens.
     */
    private static long  killJvmTimeout = TWO_MINUTES;

    /**
     * Utility class.
     */
    private Bootstrap() {

    }

    /**
     * Starts the Framework and launch threads for monitor a shutdown call.
     * @param args the arguments with the path to the conf file.
     */
    public void start(final String[] args) {


        // Read arguments in order to know where is configuration object
        if (args == null || args.length != 2) {
            logger.error("Invalid arguments");
            return;

        }

        // Get path
        String configurationPath = args[1];
        if (configurationPath == null) {
            logger.error("No configuration path found");
            return;
        }

        logger.debug("Using configuration stored in {0}", configurationPath);

        // Read configuration object from the filesystem
        File file = new File(configurationPath);
        FileInputStream fileInputStream = null;
        ObjectInputStream objectInputStream = null;
        Object readObject = null;
        try {
            fileInputStream = new FileInputStream(file);
            objectInputStream = new ObjectInputStream(fileInputStream);
            readObject = objectInputStream.readObject();
        } catch (FileNotFoundException e) {
            logger.error("Cannot find the file {0}", file, e);
            return;
        } catch (IOException e) {
            logger.error("Cannot read the file {0}", file, e);
            return;
        } catch (ClassNotFoundException e) {
            logger.error("Cannot find the Configuration class", e);
            return;
        } finally {
            // Close everything
            if (fileInputStream != null) {
                try {
                    fileInputStream.close();
                } catch (IOException e) {
                    logger.error("Cannot close stream", e);
                }
            }
            if (objectInputStream != null) {
                try {
                    objectInputStream.close();
                } catch (IOException e) {
                    logger.error("Cannot close stream", e);
                }
            }
        }

        // Ensure that object is a valid type
        if (readObject instanceof Configuration) {
            this.configuration = (Configuration) readObject;
        } else {
            logger.error("Object read on the disk is not a Configuration object. Found {0}", readObject);
            return;
        }
        //Thread that kills the jvm whatever happens.
        Thread killJvmThread = new Thread("STARTING WAITER") {

            @Override
            public void run() {
                try {
                    Thread.sleep(killJvmTimeout);
                } catch (InterruptedException e) {
                    logger.error("Cannot sleep");
                }

                // Exit JVM after timeout
                System.out.println("Timeout ! exiting");
                System.exit(1);
            }
        };
        killJvmThread.setDaemon(true);
        killJvmThread.start();

        // Set system properties
        Properties jvmProperties = configuration.getSystemProperties();
        if (jvmProperties != null) {
            Enumeration<Object> enumeration = jvmProperties.keys();
            while (enumeration.hasMoreElements()) {
                String key = enumeration.nextElement().toString();
                String value = jvmProperties.get(key).toString();
                if (key.equalsIgnoreCase(MONITORED_FILE_KEY)) {
                    //set the monitored file.
                    setMonitoredFile(MONITORED_FILE_PREFIX + value);
                } else if (key.equalsIgnoreCase(KILL_JVM_TIMEOUT)) {
                    //Set the time to wait before killing the JVM whatever happens.
                    setKillJvmTiemout(value);
                } else {
                    // property already exists ?
                    String oldValue = System.getProperty(key);
                    if (!value.equals(oldValue)) {
                        System.setProperty(key, value);
                        logger.debug("Adding System property '" + key + "' with value '" + value + "'");
                    }
                }

            }
        }
        try {
            this.frameworkManager = FrameworkManagerHelper.getFrameworkManager(configuration);
        } catch (FrameworkManagerException e) {
            logger.error("Cannot get the framework manager", e);
            System.exit(1);
        }

        // start it
        try {
            this.frameworkManager.start();
        } catch (FrameworkManagerException e) {
            logger.error("Cannot start the framework", e);
            System.exit(1);
        }


        // Launch a thread that monitor a file and stop the server if it
        // contains STOPPED_INSTRUCTION

        Thread stopDetector = new Thread("STOP DETECTOR") {

            @Override
            public void run() {
                File instructionsFile = configuration.getInstructionsFile();
                if (instructionsFile == null) {
                    instructionsFile = new File(monitoredFile);
                }
                logger.debug("Thread stopped detector scanning file ''{0}'' for instructions", instructionsFile);
                boolean stopSend = false;
                while (!stopSend) {

                    // File exists, read content
                    if (instructionsFile.exists()) {
                        FileReader fileReader;
                        try {
                            fileReader = new FileReader(instructionsFile);
                            BufferedReader bufferedReader = new BufferedReader(fileReader);
                            String instruction = bufferedReader.readLine();
                            if (STOPPED_INSTRUCTION.equals(instruction)) {
                                logger.debug("Detect STOP instruction !");
                                stopSend = true;
                                shutdown();

                            }
                            try {
                                bufferedReader.close();
                            } catch (IOException e) {
                                logger.debug("Unable to close the reader", e);
                            }
                        } catch (FileNotFoundException e) {
                            logger.debug("File not found", e);
                        } catch (IOException e) {
                            logger.debug("Cannot read file", e);
                        }
                    }

                    // Wait a little
                    try {
                        Thread.sleep(WAIT_MONITORING);
                    } catch (InterruptedException e) {
                        logger.error("Cannot sleep", e);
                    }
                }
            }
        };
        stopDetector.setDaemon(true);
        stopDetector.start();

        // exit code of the bootsrap if failures on the framework manager (like tests)
        if (configuration.isShutdownAfterStart()) {
            System.exit(frameworkManager.getExitCode());
        }
    }

    /**
     * @param timeout The time out value.
     */
    private static void setKillJvmTiemout(final String timeout) {
        killJvmTimeout = Long.parseLong(timeout) * SECONDS;
    }
    /**
     * Stop this framework.
     */
    public void shutdown() {
        if (shutdownDone) {
            return;
        }
        shutdownDone = true;
        try {
            this.frameworkManager.stop();
        } catch (FrameworkManagerException e) {
            logger.error("Cannot stop the framework", e);
        }

        logger.debug("Waiting {0} seconds before killing the JVM {1}", configuration.getWaitAfterStop(), configuration.getURLConfigFile());
        try {
            Thread.sleep(configuration.getWaitAfterStop() * SECONDS);
        } catch (InterruptedException e) {
            logger.error("Cannot wait", e);
        }
        System.exit(frameworkManager.getExitCode());
    }

    /**
     * Starts the bootstrap class. It will read the configuration from a given
     * argument.
     * @param args contains the path to the configuration object to use.
     */
    public static void main(final String[] args) {
        new Bootstrap().start(args);
    }

    /**
     * @return the monitoredFile
     */
    public static String getMonitoredFile() {
        return monitoredFile;
    }

    /**
     * @param monitoredFileParam the monitoredFile to set
     */
    public static void setMonitoredFile(final String monitoredFileParam) {
        monitoredFile = monitoredFileParam;
    }

    /**
     * return the monitored file built from the given name.
     * @param monitoredFileName the file name
     * @return url string of the file.
     */
    public static String getMonitoredFile(final String monitoredFileName) {
        //return default monitored file
        if (monitoredFileName == null) {
            return getMonitoredFile();
        }
        //build monitored file from the given name.
        return MONITORED_FILE_PREFIX  + monitoredFileName;
    }
}
