/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.osgi.launcher.core;

import java.io.File;
import java.io.Serializable;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;


/**
 * This class keeps the configuration. It is a Serializable object then this
 * object can be read from the forked JVM if any.
 * @author Florent BENOIT
 */
public class Configuration implements Serializable {

    /**
     * Serial version UID.
     */
    private static final long serialVersionUID = 6556052376136830050L;

    /**
     * Waiting value after the start.
     */
    private long waitAfterStart = 0;

    /**
     * Waiting value after the stop.
     */
    private long waitAfterStop = 0;

    /**
     * List of bundles to start.
     */
    private List<URL> bundles = null;

    /**
     * List of framework bundles to start.
     */
    private List<BundleDesc> frameworkBundles = null;

    /**
     * List of test framework bundles to start.
     */
    private List<BundleDesc> testFrameworkBundles = null;

    /**
     * List of bundles to start after a given delay.
     */
    private DelayedBundlePattern[] delayedBundles = null;

    /**
     * List of bundles to start in the given order (by their symbolic name).
     */
    private String[] startOrderBundles = null;

    /**
     * List of bundles to stop in the given order (by their symbolic name).
     */
    private String[] stopOrderBundles = null;

    /**
     * List of bundles to ignore (by their symbolic name).
     */
    private String[] excludedBundles = null;

    /**
     * List of path for the user Classpath.
     */
    private String[] userClasspath = null;

    /**
     * System properties.
     */
    private Properties systemProperties = null;

    /**
     * Working directory.
     */
    private File workingDirectory = null;

    /**
     * Config file URL.
     */
    private URL urlConfigFile = null;

    /**
     * Configuration elements.
     */
    private File[] configurationElements;

    /**
     * Instructions file.
     */
    private File instructionsFile = null;

    /**
     * OSGi Configuration (if any).
     */
    private Map<String, String> osgiConfiguration = null;

    /**
     * Framework interfaces to wait in order to be sure that framework is ready.
     */
    private List<String> frameworkWaitInterfaces = null;

    /**
     * Shutdown after a start ?
     */
    private boolean shutdownAfterStart = false;

    /**
     * Wait the end of the remote process (when forking).
     * Shouldn't be used for pre/post integration test phase
     */
    private boolean waitEndProcess = false;

    /**
     * Output directory for writing.
     */
    private File testDirectory;


    /**
     * Default constructor.
     */
    public Configuration() {
        this.systemProperties = new Properties();
        this.frameworkBundles = new ArrayList<BundleDesc>();
        this.testFrameworkBundles = new ArrayList<BundleDesc>();
        this.osgiConfiguration = new HashMap<String, String>();
        this.frameworkWaitInterfaces = new ArrayList<String>();
    }


    /**
     * @return Waiting value after the start.
     */
    public long getWaitAfterStart() {
        return waitAfterStart;
    }

    /**
     * @return List of bundles to start in the given order (by their symbolic
     *         name).
     */
    public String[] getStartOrderBundles() {
        if (startOrderBundles != null) {
            return startOrderBundles.clone();
        }
        return null;
    }

    /**
     * Sets the List of bundles to start in the given order (by their symbolic
     * name).
     * @param startOrderBundles the given list
     */
    public void setStartOrderBundles(final String[] startOrderBundles) {
        if (startOrderBundles != null) {
            this.startOrderBundles = startOrderBundles.clone();
        }
    }

    /**
     * @return List of bundles to stop in the given order (by their symbolic
     *         name).
     */
    public String[] getStopOrderBundles() {
        if (stopOrderBundles != null) {
            return stopOrderBundles.clone();
        }
        return null;
    }

    /**
     * Sets the List of bundles to stop in the given order (by their symbolic
     * name).
     * @param stopOrderBundles the given list
     */
    public void setStopOrderBundles(final String[] stopOrderBundles) {
        if (stopOrderBundles != null) {
            this.stopOrderBundles = stopOrderBundles.clone();
        }
    }

    /**
     * @return List of bundles to ignore(by their symbolic
     *         name).
     */
    public String[] getExcludedBundles() {
        if (excludedBundles != null) {
            return excludedBundles.clone();
        }
        return null;
    }

    /**
     * Sets the List of bundles to ignore (by their symbolic
     * name).
     * @param excludedBundles the given list
     */
    public void setExcludedBundles(final String[] excludedBundles) {
        if (excludedBundles != null) {
            this.excludedBundles = excludedBundles.clone();
        }
    }

    /**
     * @return List of entry to add in the classpath.
     */
    public String[] getUserClasspath() {
        if (userClasspath != null) {
            return userClasspath.clone();
        }
        return null;
    }

    /**
     * Sets the List of entries to add in the classpath (by their symbolic
     * name).
     * @param userClasspath the given list
     */
    public void setUserClasspath(final String[] userClasspath) {
        if (userClasspath != null) {
            this.userClasspath = userClasspath.clone();
        }
    }

    /**
     * Sets the List of bundles to start.
     * @param bundles the List of bundles to start.
     */
    public void setBundles(final List<URL> bundles) {
        this.bundles = bundles;
    }

    /**
     * @return the List of bundles to start.
     */
    public List<URL> getBundles() {
        return bundles;
    }

    /**
     * Sets the List of framework bundles to start.
     * @param frameworkBundles the List of framework bundles to start.
     */
    public void setFrameworkBundles(final List<BundleDesc> frameworkBundles) {
        this.frameworkBundles = frameworkBundles;
    }

    /**
     * @return the List of framework test bundles to install/start.
     */
    public List<BundleDesc> getTestFrameworkBundles() {
        return testFrameworkBundles;
    }

    /**
     * Sets the List of framework test bundles to start.
     * @param testFrameworkBundles the List of framework test bundles to start.
     */
    public void setTestFrameworkBundles(final List<BundleDesc> testFrameworkBundles) {
        this.testFrameworkBundles = testFrameworkBundles;
    }

    /**
     * @return the List of framework bundles to install/start.
     */
    public List<BundleDesc> getFrameworkBundles() {
        return frameworkBundles;
    }


    /**
     * Sets Waiting value after the start.
     * @param waitAfterStart the given value in seconds
     */
    public void setWaitAfterStart(final long waitAfterStart) {
        this.waitAfterStart = waitAfterStart;
    }

    /**
     * @return List of bundles to start after a given delay.
     */
    public DelayedBundlePattern[] getDelayedBundles() {
        if (delayedBundles != null) {
            return delayedBundles.clone();
        }
        return null;
    }

    /**
     * Sets the List of bundles to start after a given delay.
     * @param delayedBundles the given list
     */
    public void setDelayedBundles(final DelayedBundlePattern[] delayedBundles) {
        if (delayedBundles != null) {
            this.delayedBundles = delayedBundles.clone();
        }
    }

    /**
     * @return configuration file used by the framework.
     */
    public URL getURLConfigFile() {
        return urlConfigFile;
    }

    /**
     * Sets the configuration file used by the framework.
     * @param urlConfigFile URL of the config file
     */
    public void setURLConfigFile(final URL urlConfigFile) {
        this.urlConfigFile = urlConfigFile;
    }

    /**
     * Gets the working directory of the framework.
     * @return the working directory
     */
    public File getWorkingDirectory() {
        return workingDirectory;
    }

    /**
     * Sets the working directory.
     * @param workingDirectory the given directory
     */
    public void setWorkingDirectory(final File workingDirectory) {
        this.workingDirectory = workingDirectory;
    }

    /**
     * @return system properties to use for the launch of the framework
     */
    public Properties getSystemProperties() {
        return systemProperties;
    }

    /**
     * Sets the system properties to use for the launch of the framework.
     * @param systemProperties the set of properties
     */
    public void setSystemProperties(final Properties systemProperties) {
        if (systemProperties != null) {
            this.systemProperties = systemProperties;
        }
    }

    /**
     * @return Waiting value after the stop.
     */
    public long getWaitAfterStop() {
        return waitAfterStop;
    }

    /**
     * Sets the Waiting value after the stop.
     * @param waitAfterStop the wait value in seconds.
     */
    public void setWaitAfterStop(final long waitAfterStop) {
        this.waitAfterStop = waitAfterStop;
    }

    /**
     * @return instructions file used by the framework.
     */
    public File getInstructionsFile() {
        return instructionsFile;
    }

    /**
     * @param instructionFile the instructions file used by the framework.
     */
    public void setInstructionsFile(final File instructionsFile) {
        this.instructionsFile = instructionsFile;
    }

    public void setConfigurationElements(final File[] configurationElements) {
        this.configurationElements = configurationElements;
    }

    public File[] getConfigurationElements() {
        return configurationElements;
    }

    public Map<String, String> getOsgiConfiguration() {
        return osgiConfiguration;
    }


    public void setOsgiConfiguration(final Map<String, String> osgiConfiguration) {
        this.osgiConfiguration = osgiConfiguration;
    }

    public List<String> getFrameworkWaitInterfaces() {
        return frameworkWaitInterfaces;
    }


    public void setFrameworkWaitInterfaces(final List<String> frameworkWaitInterfaces) {
        this.frameworkWaitInterfaces = frameworkWaitInterfaces;
    }


    public boolean isShutdownAfterStart() {
        return shutdownAfterStart;
    }


    public void setShutdownAfterStart(final boolean shutdownAfterStart) {
        this.shutdownAfterStart = shutdownAfterStart;
    }


    public boolean isWaitEndProcess() {
        return waitEndProcess;
    }


    public void setWaitEndProcess(final boolean waitEndProcess) {
        this.waitEndProcess = waitEndProcess;
    }



    public File getTestDirectory() {
        return testDirectory;
    }

    public void setTestDirectory(final File testDirectory) {
        this.testDirectory = testDirectory;
    }

}
