/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.osgi.launcher.core;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.Constants;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.launch.Framework;
import org.osgi.framework.launch.FrameworkFactory;
import org.osgi.framework.startlevel.BundleStartLevel;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.maven.osgi.launcher.api.FrameworkException;
import org.ow2.util.maven.osgi.launcher.test.api.ITestLauncher;
import org.ow2.util.maven.osgi.launcher.test.api.ITestReport;
import org.ow2.util.substitution.IPropertyResolver;
import org.ow2.util.substitution.engine.DefaultSubstitutionEngine;
import org.ow2.util.substitution.resolver.ChainedResolver;
import org.ow2.util.substitution.resolver.MapResolver;
import org.ow2.util.substitution.resolver.PropertiesResolver;
import org.ow2.util.substitution.resolver.RecursiveResolver;
import org.ow2.util.url.URLUtils;


/**
 * @author Florent BENOIT
 */
public class FrameworkManager {

    /**
     * Seconds is 1000 millis.
     */
    private static final long SECONDS = 1000L;

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(FrameworkManager.class);

    /**
     * Configuration object.
     */
    private Configuration configuration = null;

    /**
     * Framework factory to use.
     */
    private FrameworkFactory frameworkFactory;

    /**
     * Exit code.
     */
    private int exitCode = 1;


    /**
     * Framework to use.
     */
    private Framework framework;

    private DefaultSubstitutionEngine engine = null;

    /**
     * Execute the selected framework.
     * @throws FrameworkManagerException if plugin is not executed
     */
    public void start() throws FrameworkManagerException {


        // Configure Framework
        logger.debug("Creating Framework");
        try {
            URL configUrl = configuration.getURLConfigFile();
            if (configUrl != null) {
                configure(configUrl);
            } else {
                configure(configuration.getConfigurationElements());
            }

        } catch (FrameworkException e) {
            throw new FrameworkManagerException("Cannot configure the framework", e);
        }

        // Init
        logger.info("Using OSGi Framework ''{0}'' v{1}", framework.getSymbolicName(), framework.getVersion());
        try {
            framework.init();
        } catch (BundleException e) {
            throw new FrameworkManagerException("Cannot init the framework", e);
        }


        // start the framework
        logger.debug("Starting Framework");
        try {
            framework.start();
        } catch (BundleException e) {
            throw new FrameworkManagerException("Cannot start the framework", e);
        }

        // Use the Bundle context of the framework
        BundleContext bundleContext = framework.getBundleContext();


        // Now, if there are framework bundles, launch them and wait for interfaces being exposed
        logger.debug("framework bundles = ''{0}''", configuration.getFrameworkBundles());
        installStart(bundleContext, configuration.getFrameworkBundles());
        waitFor(bundleContext, configuration.getFrameworkWaitInterfaces());

        // If there are framework test bundles, launch them
        logger.debug("Test framework bundles = ''{0}''", configuration.getTestFrameworkBundles());
        installStart(bundleContext, configuration.getTestFrameworkBundles());


        // Deploy the bundles given by the user
        logger.debug("Installing bundles");
        List<Bundle> osgiBundles = new LinkedList<Bundle>();
        // Install bundles
        for (URL bundle : configuration.getBundles()) {
            try {
                logger.debug(" -> Installing bundle {0}", bundle);
                Bundle osgiBundle = bundleContext.installBundle(bundle.toExternalForm());
                osgiBundles.add(osgiBundle);
                logger.debug("     -> Bundle is {0}", osgiBundle);
            } catch (BundleException e) {
                throw new FrameworkManagerException("Cannot install the bundle '" + bundle + "'", e);
            }
        }

        // Compute list of bundles to delayed.
        List<String> patternsDelayed = new ArrayList<String>();
        if (configuration.getDelayedBundles() != null) {
            for (DelayedBundlePattern delayedBundlePattern : configuration.getDelayedBundles()) {
                logger.debug("Adding {0}", delayedBundlePattern);
                patternsDelayed.add(delayedBundlePattern.getPattern());
            }
        }


        // Auto Start bundles in a specified order ?
        String[] orderedBundles = configuration.getStartOrderBundles();
        if (orderedBundles != null) {
            // Search the bundle of the installed bundle and if there is a
            // match, start it
            for (String symbolicName : orderedBundles) {
                for (Bundle bundle : osgiBundles) {
                    if (symbolicName.equals(bundle.getSymbolicName())) {
                        logger.debug(" -> Starting ordered bundle {0}", bundle);
                        start(bundle);
                    }
                }
            }
        }

        // Start remaining bundles
        logger.debug("Starting bundles");
        for (Bundle bundle : osgiBundles) {

            // Excluded
            if (configuration.getExcludedBundles() != null
                    && Arrays.asList(configuration.getExcludedBundles()).contains(bundle.getSymbolicName())) {
                break;
            }

            // Delayed ? wait
            boolean delayed = false;
            // Delayed bundle ?
            for (String pattern : patternsDelayed) {
                if (bundle.getSymbolicName().matches(pattern)) {
                    logger.debug("  -> Delaying start of {0}", bundle.getSymbolicName());
                    delayed = true;
                    break;
                }
            }

            // Start only if not delayed
            if (!delayed) {
                start(bundle);
            }
        }

        // Start bundle after their delayed time.
        // TODO : Launch in a thread...
        DelayedBundlePattern[] delayedBundles = configuration.getDelayedBundles();
        if (delayedBundles != null) {
            for (DelayedBundlePattern delayedBundle : delayedBundles) {
                // Wait given time
                try {
                    Thread.sleep(delayedBundle.getSecondsToWait() * SECONDS);
                } catch (InterruptedException e) {
                    throw new FrameworkManagerException("Cannot wait for delayed bundles", e);
                }

                String pattern = delayedBundle.getPattern();

                // Start all matching bundles
                for (Bundle bundle : osgiBundles) {
                    if (bundle.getSymbolicName().matches(pattern)) {
                        logger.debug(" -> Starting delayed bundle {0}", bundle);
                        start(bundle);
                    }
                }
            }
        }

        // Launch tests
        if (getConfiguration().getTestFrameworkBundles() != null && !getConfiguration().getTestFrameworkBundles().isEmpty()) {
            logger.debug("There are test framework bundles, so wait for a service");
            ServiceReference<ITestLauncher> serviceReference = waitFor(bundleContext, ITestLauncher.class);
            logger.debug("Test service found with the serviceReference ''{0}''", serviceReference);
            if (serviceReference != null) {
                ITestLauncher service = bundleContext.getService(serviceReference);
                logger.debug("Launching tests with the service ''{0}''", service);

                // configuration
                TestConfiguration config = new TestConfiguration();
                config.setOutputDirectory(getConfiguration().getTestDirectory());
                ITestReport report = service.runTests(config);
                if (report == null || !report.success()) {
                    logger.debug("Tests are not successful");
                } else if (report.success()) {
                    logger.debug("Tests are successful");
                    exitCode = 0;
                }
            }
        } else {
            // no tests
            exitCode = 0;
        }

        if (configuration.isShutdownAfterStart()) {
            stop();
        } else if (configuration.getWaitAfterStart() > 0) {
            try {
                Thread.sleep(configuration.getWaitAfterStart() * SECONDS);
            } catch (InterruptedException e) {
                throw new FrameworkManagerException("Cannot wait", e);
            }
        }



    }

    /**
     * Create and configure the framework with the given configuration files.
     * @param elements an array of File containing elements of a system description
     * @throws FrameworkException if configuration fails
     */
    private void configure(final File[] elements) throws FrameworkException {

        // Prepare the resolver
        ChainedResolver chainedResolver = new ChainedResolver();
        this.engine = new DefaultSubstitutionEngine();
        engine.setMarkerChar('$');
        engine.setOpeningChar('{');
        engine.setEndingChar('}');
        engine.setResolver(new RecursiveResolver(engine, chainedResolver));

        chainedResolver.getResolvers().add(new PropertiesResolver(System.getProperties()));

        Map<String, String> map = new HashMap<String, String>();

        // Add a special property 'javase.version'
        // Detect the JVM version (1.5 / 1.6), fallback on 1.5
        map.put("javase.version", javaSeSpecificationVersion());

        map.putAll(configuration.getOsgiConfiguration());
        logger.debug("map content = " + map);

        chainedResolver.getResolvers().add(new MapResolver(map));

        if (elements != null) {
            for (File file : elements) {

                URL element = URLUtils.fileToURL(file);

                // Open a connection on the given URL
                URLConnection urlConnection;
                try {
                    urlConnection = element.openConnection();
                } catch (IOException e) {
                    throw new FrameworkException("Cannot open the connection on the URL '" + element + "'.", e);
                }
                urlConnection.setDefaultUseCaches(false);

                // Populate a properties object with the stream of the URL
                Properties prop = new Properties();
                InputStream is = null;
                try {
                    is = urlConnection.getInputStream();
                    prop.load(is);
                } catch (IOException e) {
                    throw new FrameworkException("Cannot open the connection on the URL '" + element + "'.", e);
                } finally {
                    if (is != null) {
                        try {
                            is.close();
                        } catch (IOException e) {
                            logger.error("Cannot close the stream", e);
                        }
                    }
                }
                map.putAll(props2Map(prop));
                chainedResolver.getResolvers().add(new MapResolver(map));

            }
        }
        resolveProperties(map, chainedResolver);



        // Sets the storage framework
        map.put(Constants.FRAMEWORK_STORAGE, configuration.getWorkingDirectory().toString());

        logger.debug("Configuring framework with the properties ''{0}''", map);


        // Create framework
        this.framework = frameworkFactory.newFramework(map);

    }

    /**
     * Convert the given properties object into a map<String,String>.
     * @param p the properties object
     * @return a map
     */
    public static Map<String, String> props2Map(final Properties p) {
        Map<String, String> map = new HashMap<String, String>();
        if (p != null) {
            Iterator<Entry<Object,Object>> it = p.entrySet().iterator();
            while (it.hasNext()) {
                Entry<Object,Object> entry = it.next();
                map.put(entry.getKey().toString(), entry.getValue().toString());
            }
        }
        return map;
    }

    /**
     * Resolve the given properties with the help of the resolver.
     * @param map the properties
     * @param resolver the properties resolver
     */
    private void resolveProperties(final Map<String, String> map, final IPropertyResolver resolver) {
        Iterator<String> names = map.keySet().iterator();
        while (names.hasNext()) {
            String name = names.next();

            String value = map.get(name);
            String systemValue = System.getProperty(name);

            // Override with System properties if possible
            if (systemValue != null) {
                value = systemValue;
            }

            // Resolve the value
            String resolved = engine.substitute(value);
            map.put(name, resolved);
        }
    }


    /**
     * @return the java SE version used at runtime
     */
    private static String javaSeSpecificationVersion() {

        // Try package version
        String version = System.class.getPackage().getSpecificationVersion();
        if (version != null) {
            return version;
        }

        // Try system property
        version = System.getProperty("java.specification.version");
        if (version != null) {
            return version;
        }

        // Fallback value
        return "1.5";

    }



    /**
     * Create and configure the framework with the given configuration file.
     * @param url the path to the framework configuration
     * @throws FrameworkException if configuration fails
     */
    protected void configure(final URL url) throws FrameworkException {

        Map<String, String> mapConfig = new HashMap<String, String>();

        // Open a connection on the given URL
        URLConnection urlConnection;
        try {
            urlConnection = url.openConnection();
        } catch (IOException e) {
            throw new FrameworkException("Cannot open the connection on the URL '" + url + "'.", e);
        }
        urlConnection.setDefaultUseCaches(false);

        // Populate a properties object with the stream of the URL
        Properties prop = new Properties();
        InputStream is = null;
        try {
            is = urlConnection.getInputStream();
            prop.load(is);
        } catch (IOException e) {
            throw new FrameworkException("Cannot open the connection on the URL '" + url + "'.", e);
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (IOException e) {
                    logger.error("Cannot close the stream", e);
                }
            }
        }
        // Extract a given set of properties
        for (Entry<?, ?> propertyKey : prop.entrySet()) {
            logger.debug("Configuring property ''{0}'' with value ''{1}''", propertyKey.getKey(), propertyKey.getValue());
            mapConfig.put(propertyKey.getKey().toString(), propertyKey.getValue().toString());
        }

        // Sets the storage framework
        mapConfig.put(Constants.FRAMEWORK_STORAGE, configuration.getWorkingDirectory().toString());

        // Create framework
        this.framework = frameworkFactory.newFramework(mapConfig);

    }


    /**
     * Execute the plugin in the sub class.
     * @throws FrameworkManagerException if the framework can't be stopped
     */
    public void stop() throws FrameworkManagerException {

        // Auto stop bundles in a specified order ?
        String[] orderedBundles = configuration.getStopOrderBundles();
        if (orderedBundles != null) {

            // List of bundles managed by the framework
            Bundle[] frameworkBundles = framework.getBundleContext().getBundles();
            if (frameworkBundles != null) {

                // Search the bundle of the installed bundle and if there is a
                // match, stop it
                for (String symbolicName : orderedBundles) {

                    for (Bundle bundle : frameworkBundles) {
                        if (symbolicName.equals(bundle.getSymbolicName())) {
                            logger.info(" -> Stopping ordered bundle {0}", bundle);
                            try {
                                bundle.stop();
                            } catch (BundleException e) {
                                throw new FrameworkManagerException("Cannot stop the bundle '" + bundle + "'", e);
                            }
                        }
                    }
                }
            }
        }

        logger.debug("Stopping Framework...");
        // Stop the framework
        try {
            framework.stop();
        } catch (BundleException e) {
            throw new FrameworkManagerException("Cannot stop the framework", e);
        }

        logger.debug("Framework stopped");
    }

    /**
     * Gets the configuration used by this manager.
     * @return the configuration.
     */
    public Configuration getConfiguration() {
        return configuration;
    }

    /**
     * Sets the configuration used by this manager.
     * @param configuration the given configuration.
     */
    public void setConfiguration(final Configuration configuration) {
        this.configuration = configuration;
    }

    /**
     * Sets the framework factory that the manager will use to build framework.
     * @param frameworkFactory the given framework factory
     */
    public void setFrameworkFactory(final FrameworkFactory frameworkFactory) {
        this.frameworkFactory = frameworkFactory;
    }

    /**
     * Starts the given bundle.
     * @param bundle the given bundle
     * @throws FrameworkManagerException if the bundle is not started
     */
    public void start(final Bundle bundle) throws FrameworkManagerException {
        if (isFragmentBundle(bundle)) {
            logger.debug("Fragment bundle ''{0}'' is not started", bundle);
            return;
        }
        try {
            logger.debug(" -> Starting bundle {0}", bundle);
            bundle.start();
        } catch (BundleException e) {
            throw new FrameworkManagerException("Cannot start the bundle '" + bundle + "'", e);
        }
    }

    /**
     * @param bundle the bundle to test
     * @return true if the given bundle is a fragment bundle
     */
    protected boolean isFragmentBundle(final Bundle bundle) {
        return bundle.getHeaders().get(Constants.FRAGMENT_HOST) != null;
    }

    /**
     * Wait that the following interfaces are ready.
     * @param bundleContext the bundle context used to start and install bundles
     * @param waitingInterfaces the given interface name to wait for
     * @throws FrameworkManagerException if unable to start
     */
    protected <T> ServiceReference<T> waitFor(final BundleContext bundleContext, final Class<T> waitingInterface) throws FrameworkManagerException {
        // Do not process further, no wait
        if (waitingInterface == null) {
            return null;
        }

        boolean interfaceAvailable = false;
        int waitingCount = 0;
        ServiceReference<T> serviceReference =  null;
        // Now wait for the given interfaces
        while (!interfaceAvailable && waitingCount < 500) {
            serviceReference  = bundleContext.getServiceReference(waitingInterface);
            if (serviceReference != null) {
                logger.debug("Interface ''{0}'' is available", waitingInterface);
            }
            interfaceAvailable = true;

            try {
                Thread.sleep(100);
                waitingCount++;
            } catch (InterruptedException e) {
                // nothing
            }
        }

        if (!interfaceAvailable) {
            throw new FrameworkManagerException("Unable to find interface '" + waitingInterface + "'.");
        }

        return serviceReference;
    }

    /**
     * Wait that the following interfaces are ready.
     * @param bundleContext the bundle context used to start and install bundles
     * @param waitingInterfaces the given interface name to wait for
     * @throws FrameworkManagerException if unable to start
     */
    protected ServiceReference<?> waitFor(final BundleContext bundleContext, final String waitingInterface) throws FrameworkManagerException {

        // Do not process further, no wait
        if (waitingInterface == null) {
            return null;
        }

        boolean interfaceAvailable = false;
        int waitingCount = 0;
        ServiceReference<?> serviceReference =  null;
        // Now wait for the given interfaces
        while (!interfaceAvailable && waitingCount < 500) {
            serviceReference  = bundleContext.getServiceReference(waitingInterface);
            if (serviceReference != null) {
                logger.debug("Interface ''{0}'' is available", waitingInterface);
            }
            interfaceAvailable = true;

            try {
                Thread.sleep(100);
                waitingCount++;
            } catch (InterruptedException e) {
                // nothing
            }
        }

        if (!interfaceAvailable) {
            throw new FrameworkManagerException("Unable to find interface '" + waitingInterface + "'.");
        }

        return serviceReference;

    }



    /**
     * Wait that the following interfaces are ready.
     * @param bundleContext the bundle context used to start and install bundles
     * @param waitingInterfaces the given interface name to wait for
     * @throws FrameworkManagerException if unable to start
     */
    protected void waitFor(final BundleContext bundleContext, final List<String> waitingInterfaces)
            throws FrameworkManagerException {

        // Do not process further, no wait
        if (waitingInterfaces == null || waitingInterfaces.isEmpty()) {
            return;
        }

        boolean allInterfacesAvailable = false;
        int waitingCount = 0;

        // Now wait for the given interfaces
        while (!allInterfacesAvailable  && waitingCount < 500) {

            boolean found = true;
            for (String itf : waitingInterfaces) {
                ServiceReference<?> serviceReference = bundleContext.getServiceReference(itf);
                if (serviceReference != null) {
                    logger.debug("Interface ''{0}'' is available", itf);
                }
                found = found && (serviceReference != null);
            }

            if (found) {
                allInterfacesAvailable = true;
            }


            try {
                Thread.sleep(100);
                waitingCount++;
            } catch (InterruptedException e) {
                // nothing
            }
        }

        if (!allInterfacesAvailable) {
            throw new FrameworkManagerException("Unable to find all interfaces among '" + waitingInterfaces + "'.");
        }

    }



    /**
     * Install and start the given set of bundles.
     * @param bundleContext the bundle context used to start and install bundles
     * @param bundles the given bundles to manage
     * @throws FrameworkManagerException if unable to start
     */
    protected void installStart(final BundleContext bundleContext, final List<BundleDesc> bundles)
            throws FrameworkManagerException {

        // Do not process further
        if (bundles == null || bundles.isEmpty()) {
            return;
        }

        List<Bundle> toStartBundles = new ArrayList<Bundle>();

        // First, install all the bundles
        for (BundleDesc bundleDesc : bundles) {
            Bundle bundle;
            try {
                // is using reference, add reference prefix
                String location = null;
                if (bundleDesc.isUseReference()) {
                    location = "reference:".concat(bundleDesc.getLocation());
                } else {
                    location = bundleDesc.getLocation();
                }
                bundle = bundleContext.installBundle(location);
            } catch (BundleException e) {
                throw new FrameworkManagerException("Cannot start the bundle '" + bundleDesc + "'", e);
            }
            // Set start level
            bundle.adapt(BundleStartLevel.class).setStartLevel(bundleDesc.getStartLevel());

            if (bundleDesc.isStart()) {
                toStartBundles.add(bundle);
            }

        }


        // Now, start them (except fragments)
        for (Bundle bundle : toStartBundles) {
            start(bundle);
        }

    }

    /**
     * @return exit code
     */
    public int getExitCode() {
        return exitCode;
    }


}
