/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.osgi.launcher.core;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.URL;
import java.net.URLConnection;

import org.osgi.framework.launch.FrameworkFactory;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;



/**
 * Build a new Framework manager for a given configuration.
 * @author Florent BENOIT
 */
public final class FrameworkManagerHelper {


    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(FrameworkManagerHelper.class);

    /**
     * Jar entry for the OSGi FrameWork factory.
     */
    private static final String FRAMEWORKFACTORY_MANIFEST_ENTRY = "META-INF/services/org.osgi.framework.launch.FrameworkFactory";


    /**
     * Utility class.
     */
    private FrameworkManagerHelper() {

    }

    /**
     * Gets a framework manager for the given configuration.
     * @param configuration the given configuration
     * @return a newly framework manager
     * @throws FrameworkManagerException if the framework manager is not built.
     */
    @SuppressWarnings("unchecked")
    public static FrameworkManager getFrameworkManager(final Configuration configuration) throws FrameworkManagerException {
        // Build a new FrameWork object using reflection (to avoid a direct dependency)
        String frameworkFactoryClassname = getFrameworkFactoryClassname();
        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();

        logger.debug("Using OSGi framework factory class ''{0}''.", frameworkFactoryClassname);

        // Load the class
        Class<FrameworkFactory> frameworkFactoryClass = null;
        try {
            frameworkFactoryClass = (Class<FrameworkFactory>) classLoader.loadClass(frameworkFactoryClassname);
        } catch (ClassNotFoundException e) {
            throw new FrameworkManagerException("Cannot load the framework factory with the classname "
                    + frameworkFactoryClassname, e);
        } catch (ClassCastException e) {
            throw new FrameworkManagerException("Cannot load the framework factory with the classname "
                    + frameworkFactoryClassname, e);
        }

        // Build framework object
        FrameworkFactory frameworkFactory = null;
        try {
            frameworkFactory = frameworkFactoryClass.newInstance();
        } catch (InstantiationException e) {
            throw new FrameworkManagerException("Cannot build instance of the class " + frameworkFactoryClassname, e);
        } catch (IllegalAccessException e) {
            throw new FrameworkManagerException("Cannot build instance of the class " + frameworkFactoryClassname, e);
        }

        // Now, can build framework manager
        FrameworkManager frameworkManager = new FrameworkManager();

        // Sets the configuration
        frameworkManager.setConfiguration(configuration);
        frameworkManager.setFrameworkFactory(frameworkFactory);

        return frameworkManager;
    }

    /**
     * Gets Framework class name.
     * @return class name.
     */
    public static String getFrameworkFactoryClassname() {

        // Load the META-INF/services/org.osgi.framework.launch.FrameworkFactory entry
        URL url = Thread.currentThread().getContextClassLoader().getResource(FRAMEWORKFACTORY_MANIFEST_ENTRY);

        // no entry ? error
        if (url == null) {
            throw new IllegalArgumentException("Cannot find entry " + FRAMEWORKFACTORY_MANIFEST_ENTRY
                    + "' used to detect OSGi Framework. Add the OSGi framework as a plugin "
                    + "dependency or as a plugin framework dependency");
        }

        // Open a connection on the given URL
        URLConnection urlConnection;
        try {
            urlConnection = url.openConnection();
        } catch (IOException e) {
            throw new IllegalStateException("Cannot open the connection on the URL '" + url + "'.", e);
        }
        urlConnection.setDefaultUseCaches(false);




        // Populate a properties object with the stream of the URL
        InputStream is = null;
        Reader reader = null;
        BufferedReader bufferedReader = null;
        String factoryClassName = null;
        try {
            is = urlConnection.getInputStream();
            reader = new InputStreamReader(is);
            bufferedReader = new BufferedReader(reader);
            factoryClassName = bufferedReader.readLine().trim();
        } catch (IOException e) {
            throw new IllegalStateException("Cannot open the connection on the URL '" + url + "'.", e);
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (IOException e) {
                    throw new IllegalStateException("Cannot close the stream", e);
                }
            }
            if (bufferedReader != null) {
                try {
                    bufferedReader.close();
                } catch (IOException e) {
                    throw new IllegalStateException("Cannot close the stream", e);
                }
            }
        }

        return factoryClassName;
    }
}
