/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.osgi.launcher.core;

import java.util.Enumeration;
import java.util.Properties;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.maven.osgi.launcher.api.LauncherException;

/**
 * Launcher that is used in the same JVM.
 * @author Florent BENOIT
 */
public class LocalLauncher extends AbsLauncher {

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(LocalLauncher.class);

    /**
     * Manager of the framework.
     */
    private FrameworkManager frameworkManager = null;


    /**
     * Build a launcher with the given configuration object.
     * @param configuration the given configuration
     */
    public LocalLauncher(final Configuration configuration) {
        super(configuration);
    }

    /**
     * Start the framework manager used by this launcher.
     * @throws LauncherException if there is a failure during the start.
     */
    public void start() throws LauncherException {

        // Set system properties
        Properties jvmProperties = getConfiguration().getSystemProperties();
        if (jvmProperties != null) {
            Enumeration<Object> enumeration = jvmProperties.keys();
            while (enumeration.hasMoreElements()) {
                String key = enumeration.nextElement().toString();
                String value = jvmProperties.get(key).toString();
                // property already exists ?
                String oldValue = System.getProperty(key);
                if (!value.equals(oldValue)) {
                    System.setProperty(key, value);
                    logger.debug("Adding System property '" + key + "' with value '" + value + "'");
                }
            }
        }


        logger.debug("Launching the Local framework...");
        try {
           this.frameworkManager = FrameworkManagerHelper.getFrameworkManager(getConfiguration());
        } catch (FrameworkManagerException e) {
            throw new LauncherException("Cannot get framework manager", e);
        }


        try {
            this.frameworkManager.start();
        } catch (FrameworkManagerException e) {
            throw new LauncherException("Cannot start the framework manager", e);
        }

        if (getConfiguration().isShutdownAfterStart() && frameworkManager.getExitCode() != 0) {
            throw new IllegalStateException(
            "Remote JVM has failed with failures. Aborting. It could be some tests that have failed.");
        }

    }

    /**
     * Stop the framework manager used by this launcher.
     * @throws LauncherException if there is a failure during the stop.
     */
    public void stop() throws LauncherException {
        try {
            this.frameworkManager.stop();
        } catch (FrameworkManagerException e) {
            throw new LauncherException("Cannot stop the framework manager", e);
        }
        // Waiting for shutdown
        logger.debug("Waiting ''{0}'' seconds for shutdown", getConfiguration().getWaitAfterStop());
        try {
            Thread.sleep(getConfiguration().getWaitAfterStop() * SECONDS);
        } catch (InterruptedException e) {
            throw new LauncherException("Cannot wait", e);
        }
    }


}
