/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.osgi.launcher.core;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.net.URL;

import org.codehaus.plexus.util.cli.CommandLineException;
import org.codehaus.plexus.util.cli.Commandline;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.maven.osgi.launcher.api.LauncherException;

/**
 * @author Florent BENOIT
 */
public class RemoteLauncher extends AbsLauncher {

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(RemoteLauncher.class);

    /**
     * URLs of the classpath of the JVM to start.
     */
    private URL[] classpathURLs = null;

    /**
     * Process of the remote JVM.
     */
    private Process process = null;

    /**
     * Debug the forked JVM for the start of the framework ? (default is false)
     * @parameter
     * @optional
     */
    private Debug debug;

    /**
     * Build a launcher with the given configuration object and the classpath
     * URLs used to create the forked JVM.
     * @param classpathURLs the URLs to use for the classpath of the JVM.
     * @param configuration the given configuration
     * @param debug debug options
     */
    public RemoteLauncher(final URL[] classpathURLs, final Configuration configuration, final Debug debug) {
        super(configuration);
        this.debug = debug;
        if (classpathURLs != null) {
            this.classpathURLs = classpathURLs.clone();
        }
    }

    /**
     * Create a new command line that will launch a Bootstrap object on a remote JVM.
     * @throws LauncherException if the starting fails.
     */
    public void start() throws LauncherException {

        // Create command line
        Commandline commandline = new Commandline();

        // Set the java command to use
        String javacommand = System.getProperty("java.home") + File.separator + "bin" + File.separator + "java";
        if (!new File(javacommand).isFile()) {
            javacommand += ".exe";
        }
        if (!new File(javacommand).isFile()) {
            throw new LauncherException("Cannot find the Java executable: " + javacommand);
        }
        // BUG: The Commandline class doesn't handle quoting properly
        if (javacommand.contains(" ")) {
            if (javacommand.contains("Program Files")) {
                javacommand = javacommand.replace("Program Files", "PROGRA~1");
            }

            if (javacommand.contains(" ")) {
                // Command has a space, warn since Commandline has bugs with this
                logger.warn("Command \"" + javacommand + "\" contains spaces. Launch will probably fail.");
            }
        }
        commandline.setExecutable(javacommand);


        // Store the instructions file
        String monitoredFileName = getConfiguration().getSystemProperties().getProperty(Bootstrap.MONITORED_FILE_KEY);
        String monitoredFile = Bootstrap.getMonitoredFile(monitoredFileName);
        File instructionFile = new File(monitoredFile);
        getConfiguration().setInstructionsFile(instructionFile);

        // Starting...
        writeInstruction(Bootstrap.STARTED_INSTRUCTION);

        // Build classpath
        String classpath = null;
        StringBuilder classpathStringBuilder = new StringBuilder();

        // Add user Classpath
        String[] userClasspath = getConfiguration().getUserClasspath();
        if (userClasspath != null) {
            for (String path : userClasspath) {
                if (classpathStringBuilder.length() > 0) {
                    classpathStringBuilder.append(File.pathSeparatorChar);
                }
                classpathStringBuilder.append(path);
            }
        }

        // Add URL classpath of the classloader
        for (URL path : classpathURLs) {
            if (classpathStringBuilder.length() > 0) {
                classpathStringBuilder.append(File.pathSeparatorChar);
            }
            classpathStringBuilder.append(path.getPath());
        }
        classpath = classpathStringBuilder.toString();

        // Wait has to be done here and not in a remote process
        long waitAfterStart = getConfiguration().getWaitAfterStart();
        if (waitAfterStart > 0) {
            getConfiguration().setWaitAfterStart(0L);
        }


        // Dump the configuration in a file
        File configFile = null;
        try {
            configFile = File.createTempFile(System.getProperty("user.name"), ".configuration");
        } catch (IOException e) {
            throw new LauncherException("Cannot create temp file", e);
        }
        FileOutputStream fileOutputStream = null;
        ObjectOutputStream objectOutputStream = null;
        try {
            fileOutputStream = new FileOutputStream(configFile);
            objectOutputStream = new ObjectOutputStream(fileOutputStream);
            objectOutputStream.writeObject(getConfiguration());
        } catch (FileNotFoundException e) {
            throw new LauncherException("Cannot write the configuration object in the file '" + configFile + "'.", e);
        } catch (IOException e) {
            throw new LauncherException("Cannot write the configuration object in the file '" + configFile + "'.", e);
        } finally {
            // Close streams
            if (fileOutputStream != null) {
                try {
                    fileOutputStream.close();
                } catch (IOException e) {
                    logger.error("Cannot close the stream", e);
                }
            }
            if (objectOutputStream != null) {
                try {
                    objectOutputStream.close();
                } catch (IOException e) {
                    logger.error("Cannot close the stream", e);
                }
            }
        }

        if (debug.isEnabled()) {
            // Sets debug opts
            commandline.createArgument().setValue("-Xdebug");
            commandline.createArgument().setValue("-Xnoagent");
            commandline.createArgument().setValue("-Djava.compiler=NONE");
            char suspendChar = 'y';
            if (!debug.isSuspend()) {
                suspendChar = 'n';
            }
            commandline.createArgument().setValue(
                    "-Xrunjdwp:transport=dt_socket,server=y,suspend=" + suspendChar + ",address=" + debug.getPort());
        }

        // Sets the classpath
        commandline.createArgument().setValue("-classpath");
        commandline.createArgument().setValue(classpath);

        // Class to launch
        commandline.createArgument().setValue(Bootstrap.class.getName());

        // Path to the configuration file with a system property
        commandline.createArgument().setValue(Bootstrap.CONFIGURATION_PATH);
        commandline.createArgument().setValue(configFile.getAbsolutePath());

        logger.debug("Command line is: " + commandline.toString());
        logger.debug("Launching the Remote JVM...");
        try {
            process = commandline.execute();
        } catch (CommandLineException e) {
            throw new LauncherException("Cannot execute the command line " + commandline.toString(), e);
        }
        logger.debug("Remote JVM started");


        // Trace the start execution
        InputStream error = process.getErrorStream();
        InputStream out = process.getInputStream();
        Thread errThread = new Thread(new ProcessLog(error, true));
        Thread outThread = new Thread(new ProcessLog(out, false));
        outThread.start();
        errThread.start();


        if (getConfiguration().isWaitEndProcess()) {
            logger.debug("Waiting the end of the process....");
            try {
                process.waitFor();
            } catch (InterruptedException e1) {
                // TODO Auto-generated catch block
                e1.printStackTrace();
            }
            logger.debug("Process has been finished with exit value ''{0}''", process.exitValue());
            if (process.exitValue() != 0) {
                throw new IllegalStateException(
                "Remote JVM has failed with failures. Aborting. It could be some tests that have failed.");
            }
        }


        // Waiting is done here and not in the remote process
        logger.debug("Waiting for " + waitAfterStart + " seconds");
        try {
            Thread.sleep(waitAfterStart * SECONDS);
        } catch (InterruptedException e) {
            throw new LauncherException("Cannot wait", e);
        }

        logger.debug("Wait finished");
    }

    /**
     * Write a stop instruction in a file in order to tell to the remote JVM that the remote server needs to be stopped.
     * @throws LauncherException if the call is not done.
     */
    public void stop() throws LauncherException {
        logger.info("Writing shutdown instruction in the configuration file");
        writeInstruction(Bootstrap.STOPPED_INSTRUCTION);
        // Waiting for shutdown
        logger.info("Waiting for shutdown");
        try {
            Thread.sleep(getConfiguration().getWaitAfterStop() * SECONDS);
        } catch (InterruptedException e) {
            throw new LauncherException("Cannot wait", e);
        }
    }

    /**
     * Writes the given instruction in a monitored file.
     * @param instruction the instruction to write
     * @throws LauncherException if the instruction is not written
     */
    protected void writeInstruction(final String instruction) throws LauncherException {
        // Write in the monitored file
        FileWriter fileWriter = null;

        File file = getConfiguration().getInstructionsFile();

        try {

            // Create directories first (if needed)
            if (!file.getParentFile().exists()) {
                file.getParentFile().mkdirs();
            }

            logger.debug("Writing {0} in {1}", instruction, file);
            fileWriter = new FileWriter(file);
            fileWriter.write(instruction);
        } catch (IOException e) {
            throw new LauncherException("Cannot write in file '" + file + "'", e);
        } finally {
            if (fileWriter != null) {
                try {
                    fileWriter.close();
                } catch (IOException e) {
                    logger.error("Cannot close file writer", e);
                }
            }
        }

    }

}
