/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.osgi.launcher.core;

import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugins.annotations.LifecyclePhase;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import org.ow2.util.maven.osgi.launcher.api.LauncherException;


/**
 * Start an OSGi framework.
 * @author Florent BENOIT
 * @since 2.0.0
 */
@Mojo(name = "start",
      defaultPhase = LifecyclePhase.PRE_INTEGRATION_TEST)
public class StartMojo extends AbsMojo {

    /**
     * Waiting value after the start.
     * @since 2.0.0
     */
    @Parameter(defaultValue = "0")
    private long waitAfterStart = 0;

    /**
     * List of bundles to start in the given order (by their symbolic name).
     * @since 2.0.0
     */
    @Parameter
    private String[] startOrderBundles = null;

    /**
     * List of bundles to start after a given delay.
     * @since 2.0.0
     */
    @Parameter
    private DelayedBundlePattern[] delayedBundles = null;

    /**
     * @param startOrderBundles the startOrderBundles to set
     */
    public void setStartOrderBundles(final String[] startOrderBundles) {
        if (startOrderBundles != null) {
            this.startOrderBundles = startOrderBundles.clone();
        }
    }

    /**
     * @param delayedBundles the delayedBundles to set
     */
    public void setDelayedBundles(final DelayedBundlePattern[] delayedBundles) {
        if (delayedBundles != null) {
            this.delayedBundles = delayedBundles.clone();
        }
    }

    /**
     * Start the OSGi framework.
     * @throws MojoExecutionException if there is a failure.
     */
    @Override
    protected void doExecute() throws MojoExecutionException {

        // Launch the framework with the given configuration
        try {
            getLauncher().start();
        } catch (LauncherException e) {
            throw new MojoExecutionException("Cannot launch the framework", e);
        }
    }

    /**
     * Gets the configuration used by the launcher by updating it with specific values.
     * @return a configuration object
     * @throws MojoExecutionException if the configuration can't be built.
     */
    @Override
    protected Configuration getConfiguration() throws MojoExecutionException {
        // Get configuration object.
        Configuration configuration = super.getConfiguration();

        // Add properties
        configuration.setWaitAfterStart(waitAfterStart);
        configuration.setStartOrderBundles(startOrderBundles);
        configuration.setDelayedBundles(delayedBundles);

        return configuration;

    }


}
