/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.osgi.fwkslauncher.core;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.repository.DefaultRepositoryRequest;
import org.apache.maven.artifact.repository.RepositoryRequest;
import org.apache.maven.artifact.resolver.ArtifactResolutionRequest;
import org.apache.maven.artifact.resolver.ArtifactResolutionResult;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.model.Dependency;
import org.apache.maven.model.DependencyManagement;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugins.annotations.Component;
import org.apache.maven.plugins.annotations.Parameter;
import org.apache.maven.project.MavenProject;
import org.apache.maven.repository.RepositorySystem;
import org.ow2.util.maven.osgi.launcher.core.Configuration;
import org.ow2.util.maven.osgi.launcher.core.Module;
import org.ow2.util.url.URLUtils;

/**
 * Abstract Mojo.
 * @author eyindanga
 */
public abstract class AbsMojo extends AbstractMojo {

    @Component
    protected MavenSession mavenSession;

    @Component
    protected RepositorySystem repositorySystem;

    @Component
    private MavenProject project;

    /**
     * Seconds is 1000 millis.
     */
    public static final long SECONDS = 1000L;

    /**
     * Time to wait for frameworks stop.
     */
    protected static long waitFrameworksStop;

    /**
     * Bundle file extension array.
     */
    private static final String[] BUNDLE_FILE_EXTENSIONS = new String[]{".jar"};

    /**
     * True if mojo is already initialized.
     */
    private static boolean initDone = false;

    /**
     * @since 2.0.0
     */
    @Parameter(required = true)
    private List<Framework> frameworks;

    /**
     * Launchers list.
     */
    protected static List<IFwksLauncher> launchers = new LinkedList<IFwksLauncher>();

    /*
     * (non-Javadoc)
     * @see org.apache.maven.plugin.AbstractMojo#execute()
     */
    public void execute() throws MojoExecutionException {
        // Initialize if necessary.
        if (!initDone) {
            init(frameworks);
        } else {
            getLog().debug("Initialization already done.");
        }
        try {
            doExecute();
        } catch (Exception e) {
            getLog().error(e);
            launchers = null;
            frameworks = null;
        }
    }

    /**
     * Initializes the mojo.
     * @param frameworks the frameworks' configurations.
     * @throws MojoExecutionException any.
     */
    private void init(final List<Framework> frameworks) throws MojoExecutionException {
        for (Iterator<Framework> iterator = frameworks.iterator(); iterator.hasNext();) {
            Framework framework = iterator.next();
            try {
                calculateWaitFrameworkStop(framework.getWaitAfterStop());
                resolveArtifacts(framework);
                extractBundlesFromArtifacts(framework.getBundles(), framework.getArtifacts());
                // Extract bundles from the given directories
                extractBundlesFromDirectories(framework.getBundleDirectories(), framework.getBundles());
                // register launcher for this framework
                registerFrameworkLauncher(framework);
            } catch (Exception e) {
                throw new MojoExecutionException(e.getMessage(), e);
            }
        }
        initDone = true;
    }

    /**
     * Compute time to wait for all frameworks stop.
     * @param waitAfterStop value to calculate from
     */
    private static void calculateWaitFrameworkStop(final long waitAfterStop) {
        if (waitAfterStop > waitFrameworksStop) {
            waitFrameworksStop = waitAfterStop;
        }

    }

    /**
     * Register launcher for the given framework.
     * @param framework the framework to register launcher for.
     * @throws MojoExecutionException any.
     */
    private static void registerFrameworkLauncher(final Framework framework) throws MojoExecutionException {
        IFwksLauncher launcher = createLauncher(framework);
        launchers.add(launcher);
    }

    /**
     * Build a new launcher (local or remote) depending of the fork mode.
     * @param framework the framework's configuration.
     * @return a launcher.
     * @throws MojoExecutionException if no launcher is created
     */
    static IFwksLauncher createLauncher(final Framework framework) throws MojoExecutionException {
        IFwksLauncher launcher = null;
        // ForK ? create a remote launcher
        if (framework.isFork()) {
            // Extract URLs of the current classloader
            // Get Method
            ClassLoader cl = Thread.currentThread().getContextClassLoader();
            Method getURLsMethod = null;
            try {
                getURLsMethod = URLClassLoader.class.getDeclaredMethod("getURLs");
            } catch (SecurityException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } catch (NoSuchMethodException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            }

            // Invoke method
            URL[] urls = null;
            boolean previousAccessible = getURLsMethod.isAccessible();
            getURLsMethod.setAccessible(true);
            try {
                urls = (URL[]) getURLsMethod.invoke(cl);
            } catch (IllegalArgumentException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } catch (IllegalAccessException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } catch (InvocationTargetException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } finally {
                getURLsMethod.setAccessible(previousAccessible);
            }

            launcher = new FwksRemoteLauncher(urls, getConfiguration(framework), framework.getDebug());
        } else {
            launcher = new FwksLocalLauncher(getConfiguration(framework));
        }

        return launcher;
    }

    /**
     * Get the configuration used by the launcher/manager.
     * @param framework Gets configuration of this framework.
     * @return a configuration object
     * @throws MojoExecutionException if the configuration can't be built.
     */
    protected static Configuration getConfiguration(final Framework framework) throws MojoExecutionException {
        Configuration configuration = new Configuration();

        // Time to wait after the stop
        configuration.setWaitAfterStop(framework.getWaitAfterStop());

        // Time to wait after the stop
        configuration.setWaitAfterStart(framework.getWaitAfterStart());

        // Stop bundles in a given order
        configuration.setStopOrderBundles(framework.getStopOrderBundles());

        // Start bundles in a given order
        configuration.setStartOrderBundles(framework.getStartOrderBundles());

        // Delayed bundles patterns.
        configuration.setDelayedBundles(framework.getDelayedBundles());

        // Ignore bundles
        configuration.setExcludedBundles(framework.getExcludedBundles());

        // User classpath
        configuration.setUserClasspath(framework.getUserClasspath());

        // JVM properties
        configuration.setSystemProperties(framework.getJvmProperties());

        // User classpath
        configuration.setUserClasspath(framework.getUserClasspath());

        // Bundles to deploy
        configuration.setBundles(framework.getBundles());

        // Set working directory
        configuration.setWorkingDirectory(new File(framework.getBuildDirectory(), framework.getName()));

        // Config file URL
        if (framework.getConfigurationFile() != null) {
            configuration.setURLConfigFile(URLUtils.fileToURL(framework.getConfigurationFile()));
        }

        return configuration;
    }

    /**
     * Execute the Mojo.
     * @throws Exception any.
     */
    protected abstract void doExecute() throws Exception;

    /**
     * Found artifacts for modules specified by the user.
     * @param framework the framework to resolve artifact for.
     * @throws MojoExecutionException if artifact resolution fails
     */
    @SuppressWarnings("unchecked")
    protected void resolveArtifacts(final Framework framework) throws MojoExecutionException {
        // No module, break here
        if (framework.getModules() == null) {
            return;
        }

        // for each module, build artifact.
        for (Module module : framework.getModules()) {
            String groupId = module.getGroupId();
            String artifactId = module.getArtifactId();
            String type = module.getType();
            String version = module.getVersion();
            String classifier = module.getClassifier();

            if (groupId == null) {
                throw new MojoExecutionException("No groupdId set for module '" + module + "'");
            }
            if (artifactId == null) {
                throw new MojoExecutionException("No artifactId set for module '" + module + "'");
            }

            if (type == null) {
                type = "jar";
                module.setType(type);
            }

            // No version
            if (version == null) {
                // try to find version with dependency management
                DependencyManagement dependencyManagement = framework.getProject().getDependencyManagement();

                if (dependencyManagement != null) {
                    // try to find a version for this component
                    List<Dependency> dependencies = dependencyManagement.getDependencies();
                    if (dependencies != null) {
                        for (Dependency dependency : dependencies) {
                            if (groupId.equals(dependency.getGroupId()) && artifactId.equals(dependency.getArtifactId())
                                    && type.equals(dependency.getType())) {
                                version = dependency.getVersion();
                                break;
                            }
                        }
                    }
                }
                if (version == null) {
                    throw new MojoExecutionException("No version set for module '" + module + "'");
                }
            }

            Artifact artifact = repositorySystem.createArtifactWithClassifier(groupId, artifactId, version, type,
                    classifier);

            // now resolve this artifact
            Artifact resolved = resolveArtifact(artifact);

            // Add artifact
            framework.getArtifacts().add(resolved);
        }
    }


    private Artifact resolveArtifact(Artifact artifact) throws MojoExecutionException {

        RepositoryRequest repositoryRequest = DefaultRepositoryRequest.getRepositoryRequest(mavenSession, project);
        ArtifactResolutionRequest request = new ArtifactResolutionRequest(repositoryRequest);
        request.setArtifact(artifact);

        // Resolve
        ArtifactResolutionResult result = repositorySystem.resolve(request);
        if (!result.isSuccess()) {
            for (Exception e : result.getExceptions()) {
                getLog().error(e);
            }

            throw new MojoExecutionException("Couldn't resolve artifact " + artifact);
        }
        return result.getArtifacts().iterator().next();
    }
    /**
     * Add the bundles that can be found in the given directory.
     * @param directory the directory to scan
     * @param bundles The bundles to add.
     */
    protected static void addBundleFromDirectory(final File directory, final List<URL> bundles) {
        if (directory.isDirectory()) {
            File[] files = directory.listFiles();
            for (File f : files) {
                if (f.isDirectory()) {
                    addBundleFromDirectory(f, bundles);
                } else {
                    String fileName = f.getName().toLowerCase();
                    boolean validExtension = false;
                    int i = 0;
                    while (!validExtension && i < BUNDLE_FILE_EXTENSIONS.length) {
                        if (fileName.endsWith(BUNDLE_FILE_EXTENSIONS[i])) {
                            validExtension = true;
                        }
                        i++;
                    }
                    if (validExtension) {
                        bundles.add(URLUtils.fileToURL(f));
                    }
                }
            }
        }
    }

    /**
     * Extract the bundles from the given artifacts.
     * @param bundles The bundles to extract.
     * @param artifacts artifacts list.
     * @throws MojoExecutionException any.
     */
    protected static void extractBundlesFromArtifacts(final List<URL> bundles, final List<Artifact> artifacts)
            throws MojoExecutionException {
        // Use URL of artifacts
        try {
            for (Artifact artifact : artifacts) {
                bundles.add(URLUtils.fileToURL(artifact.getFile()));
            }
        } catch (Exception e) {
            throw new MojoExecutionException(e.getMessage(), e);
        }
    }

    /**
     * Extract the bundles from the given directories.
     * @param bundleDirectories bundle directories.
     * @param bundles The bundles
     * @throws MojoExecutionException any.
     */
    protected static void extractBundlesFromDirectories(final File[] bundleDirectories, final List<URL> bundles)
            throws MojoExecutionException {
        // No directories, do nothing
        if (bundleDirectories == null) {
            return;
        }
        try {
            // Take bundles of the directories
            for (File directory : bundleDirectories) {
                addBundleFromDirectory(directory, bundles);
            }
        } catch (Exception e) {
            throw new MojoExecutionException(e.getMessage(), e);
        }

    }

}
