/**
 * Copyright 2008-2012 Bull S.A.S.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.util.maven.osgi.fwkslauncher.core;

import java.io.File;
import java.net.URL;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.project.MavenProject;
import org.ow2.util.maven.osgi.launcher.core.Debug;
import org.ow2.util.maven.osgi.launcher.core.DelayedBundlePattern;
import org.ow2.util.maven.osgi.launcher.core.Module;

/**
 * The framework's configuration.
 * @author eyindanga
 */
public class Framework {

    /**
     * Path of the build directory.
     */
    private String name;

    /**
     * The Id of the frame work.e.g <code>felix1</code>.
     */
    private String id = null;

    /**
     * Path of the build directory.
     */
    private File buildDirectory;

    /**
     * The maven project.
     */
    private MavenProject project;

    /**
     * Local Maven repository used by the Artifact Resolver.
     */
    private ArtifactRepository localRepository;

    /**
     * List of Maven repositories (remote) used by the Artifact Resolver.
     */
    private List<?> repositories;

    /**
     * List of directories that contains bundles.
     */
    private File[] bundleDirectories = null;

    /**
     * Fork the JVM for the start of the framework ? (default is false)
     */
    private boolean fork = false;

    /**
     * Debug the forked JVM for the start of the framework ? (default is false)
     */
    private Debug debug = new Debug();

    /**
     * List of Modules that needs to be installed/started.
     */
    private Module[] modules;

    /**
     * Configuration file.
     */
    private File configurationFile = null;

    /**
     * JVM Properties (System properties to set).
     */
    private Properties jvmProperties;

    /**
     * List of bundles to stop in the given order (by their symbolic name).
     */
    private String[] stopOrderBundles = null;

    /**
     * List of bundles to ignore (by their symbolic name).
     */
    private String[] excludedBundles = null;

    /**
     * List of path for the user Classpath.
     */
    private String[] userClasspath = null;

    /**
     * List of bundles to start.
     */
    private List<URL> bundles = new LinkedList<URL>();

    /**
     * List of Artifacts to Start.
     */
    private List<Artifact> artifacts = new LinkedList<Artifact>();

    /**
     * Waiting value after the stop.
     */
    private long waitAfterStop = 0;

    /**
     * Waiting value after the start.
     */
    private long waitAfterStart = 0;

    /**
     * List of bundles to start in the given order (by their symbolic name).
     */
    private String[] startOrderBundles = null;

    /**
     * List of bundles to start after a given delay.
     */
    private DelayedBundlePattern[] delayedBundles = null;

    /**
     * Default constructor.
     */
    public Framework() {

    }

    /**
     * @return the Name
     */
    public String getName() {
        return name;
    }

    /**
     * @param name the name to set
     */
    public void setName(final String name) {
        this.name = name;
    }

    /**
     * @return the id
     */
    public String getId() {
        return id;
    }

    /**
     * @param id the id to set
     */
    public void setId(final String id) {
        this.id = id;
    }

    /**
     * @return the waitAfterStart
     */
    public long getWaitAfterStart() {
        return waitAfterStart;
    }

    /**
     * @param waitAfterStart the waitAfterStart to set
     */
    public void setWaitAfterStart(final long waitAfterStart) {
        this.waitAfterStart = waitAfterStart;
    }

    /**
     * @return the startOrderBundles
     */
    public String[] getStartOrderBundles() {
        return startOrderBundles;
    }

    /**
     * @param startOrderBundles the startOrderBundles to set
     */
    public void setStartOrderBundles(final String[] startOrderBundles) {
        this.startOrderBundles = startOrderBundles;
    }

    /**
     * @return the delayedBundles
     */
    public DelayedBundlePattern[] getDelayedBundles() {
        return delayedBundles;
    }

    /**
     * @param delayedBundles the delayedBundles to set
     */
    public void setDelayedBundles(final DelayedBundlePattern[] delayedBundles) {
        this.delayedBundles = delayedBundles;
    }

    /**
     * @return the buildDirectory
     */
    public File getBuildDirectory() {
        return buildDirectory;
    }

    /**
     * @param buildDirectory the buildDirectory to set
     */
    public void setBuildDirectory(final File buildDirectory) {
        this.buildDirectory = buildDirectory;
    }

    /**
     * @return the project
     */
    public MavenProject getProject() {
        return project;
    }

    /**
     * @param project the project to set
     */
    public void setProject(final MavenProject project) {
        this.project = project;
    }

    /**
     * @return the localRepository
     */
    public ArtifactRepository getLocalRepository() {
        return localRepository;
    }

    /**
     * @param localRepository the localRepository to set
     */
    public void setLocalRepository(final ArtifactRepository localRepository) {
        this.localRepository = localRepository;
    }

    /**
     * @return the repositories
     */
    public List<?> getRepositories() {
        return repositories;
    }

    /**
     * @param repositories the repositories to set
     */
    public void setRepositories(final List<?> repositories) {
        this.repositories = repositories;
    }

    /**
     * @return the bundleDirectories
     */
    public File[] getBundleDirectories() {
        return bundleDirectories;
    }

    /**
     * @param bundleDirectories the bundleDirectories to set
     */
    public void setBundleDirectories(final File[] bundleDirectories) {
        this.bundleDirectories = bundleDirectories;
    }

    /**
     * @return the fork
     */
    public boolean isFork() {
        return fork;
    }

    /**
     * @param fork the fork to set
     */
    public void setFork(final boolean fork) {
        this.fork = fork;
    }

    /**
     * @return the debug
     */
    public Debug getDebug() {
        return debug;
    }

    /**
     * @param debug the debug to set
     */
    public void setDebug(final Debug debug) {
        this.debug = debug;
    }

    /**
     * @return the modules
     */
    public Module[] getModules() {
        return modules;
    }

    /**
     * @param modules the modules to set
     */
    public void setModules(final Module[] modules) {
        this.modules = modules;
    }

    /**
     * @return the configurationFile
     */
    public File getConfigurationFile() {
        return configurationFile;
    }

    /**
     * @param configurationFile the configurationFile to set
     */
    public void setConfigurationFile(final File configurationFile) {
        this.configurationFile = configurationFile;
    }

    /**
     * @return the jvmProperties
     */
    public Properties getJvmProperties() {
        return jvmProperties;
    }

    /**
     * @param jvmProperties the jvmProperties to set
     */
    public void setJvmProperties(final Properties jvmProperties) {
        this.jvmProperties = jvmProperties;
    }

    /**
     * @return the stopOrderBundles
     */
    public String[] getStopOrderBundles() {
        return stopOrderBundles;
    }

    /**
     * @param stopOrderBundles the stopOrderBundles to set
     */
    public void setStopOrderBundles(final String[] stopOrderBundles) {
        this.stopOrderBundles = stopOrderBundles;
    }

    /**
     * @return the excludedBundles
     */
    public String[] getExcludedBundles() {
        return excludedBundles;
    }

    /**
     * @param excludedBundles the excludedBundles to set
     */
    public void setExcludedBundles(final String[] excludedBundles) {
        this.excludedBundles = excludedBundles;
    }

    /**
     * @return the userClasspath
     */
    public String[] getUserClasspath() {
        return userClasspath;
    }

    /**
     * @param userClasspath the userClasspath to set
     */
    public void setUserClasspath(final String[] userClasspath) {
        this.userClasspath = userClasspath;
    }

    /**
     * @return the bundles
     */
    public List<URL> getBundles() {
        return bundles;
    }

    /**
     * @param bundles the bundles to set
     */
    public void setBundles(final List<URL> bundles) {
        this.bundles = bundles;
    }

    /**
     * @return the artifacts
     */
    public List<Artifact> getArtifacts() {
        return artifacts;
    }

    /**
     * @param artifacts the artifacts to set
     */
    public void setArtifacts(final List<Artifact> artifacts) {
        this.artifacts = artifacts;
    }

    /**
     * @return the waitAfterStop
     */
    public long getWaitAfterStop() {
        return waitAfterStop;
    }

    /**
     * @param waitAfterStop the waitAfterStop to set
     */
    public void setWaitAfterStop(final long waitAfterStop) {
        this.waitAfterStop = waitAfterStop;
    }

}
